/*******************************************************************************
**
**  Module  : tpro_drv26_2_6_13.c
**  Date    : 03/31/06
**  Purpose : This driver provides an interface to the TPRO-PCI
**
**  Copyright(C) 2006 Spectracom Corporation. All Rights Reserved.
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**
********************************************************************************
**
**     08/10/12 Added PMC/cPCI support
**     05/28/07 Modifications for multi-processor operation.
**     01/23/07 Added support for multiple simultaneous users
**     03/31/06 Updated from Linux (2.4.24) to Linux (2.6.11.12)
**
*******************************************************************************/
#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef MODULE
#  define MODULE
#endif

#include <linux/version.h>

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
#include <linux/config.h>
#include <asm/uaccess.h>
#else
#include <linux/uaccess.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif

#endif

#ifdef CONFIG_SMP
#define __SMP__
#endif

#define TPRO_DEBUG 1

/*******************************************************************************
**          Includes
*******************************************************************************/
#include <linux/device.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/pci.h>

#include <linux/sched.h>

#include <linux/interrupt.h>
#include <linux/cdev.h>
#include <linux/fs.h>

#include "../h/ddtpro.h"
#include "tpropci.h"
#include "tpro_func.h"
#include "os_wait.h"

/*******************************************************************************
**          Defines
*******************************************************************************/
#define VERSION                   "4.00"
#define IO_BASE_ADDR_REG          (3)
#define IO_BASE_ADDR_REG_32       (1)
#define MEMORY_BASE_ADDR_REG_CPCI (2)

#define CLASS_NAME        "tpropci"

/*******************************************************************************
**          Kernel Module Description
*******************************************************************************/
MODULE_AUTHOR("Spectracom Corporation (spectracomcorp.com)");
MODULE_DESCRIPTION("Spectracom TPRO/TSAT-PCI Timing Board");
MODULE_SUPPORTED_DEVICE("PCI Timing Board");
MODULE_LICENSE("GPL");

/*******************************************************************************
**          Brief Identification Strings
*******************************************************************************/
char tpropci_driver_name[]    = TPRO_PCI_DRV_NAME;
char tpropci_driver_string[]  = TPRO_PCI_DRV_STRING;
char tpropci_driver_version[] = VERSION;
char tpropci_copyright[]      = TPRO_PCI_DRV_COPYRIGHT;

/*******************************************************************************
**          Local Function Prototypes
*******************************************************************************/

static void           tpropci_cleanup     (void);

static int            tpropci_close       (struct inode *inode,
                                           struct file *filp);

static tpropci_dev_t* tpropci_find_device (struct inode *inode);

static int            tpropci_init        (void);

static int            tpropci_init_one    (struct pci_dev             *pdev,
                                           const struct pci_device_id *ent);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35)
static long           tpropci_ioctl        (struct file   *filp,
                                           unsigned int  cmd,
                                           unsigned long arg);
#else
static int            tpropci_ioctl        (struct inode  *inode,
                                           struct file   *filp,
                                           unsigned int  cmd,
                                           unsigned long arg);
#endif

static int            tpropci_open        (struct inode *inode,
                                           struct file *filp);

static void           tpropci_remove_one  (struct pci_dev *pdev);


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#define ISR_ARGUMENT_LIST int irq, void *dev_id, struct pt_regs *regs
#else
#define ISR_ARGUMENT_LIST int irq, void *dev_id
#endif

static irqreturn_t    tpropci_isr(ISR_ARGUMENT_LIST);

/*******************************************************************************
**          Typedef's and Structures
*******************************************************************************/

/*
** PCI Device table
*/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35)
static struct pci_device_id tpropci_pci_tbl[] __refdata =
#else
static struct pci_device_id tpropci_pci_tbl[] __initdata =
#endif
{
    {
        vendor:      TPRO_VENDOR_ID,
        device:      TPRO_5V_DEVICE_ID,
        subvendor:   TPRO_SUBSYSTEM_VENDOR_ID,
        subdevice:   PCI_ANY_ID,
        class:       0,
        class_mask:  0,
        driver_data: 0
    },
    {
        vendor:      TPRO_VENDOR_ID,
        device:      TPRO_3V_DEVICE_ID,
        subvendor:   TPRO_SUBSYSTEM_VENDOR_ID,
        subdevice:   PCI_ANY_ID,
        class:       0,
        class_mask:  0,
        driver_data: 0
    },
    {
        vendor:      TPRO_U66_VENDOR_ID,
        device:      TPRO_U66_DEVICE_ID,
        subvendor:   TPRO_U66_SUBSYSTEM_VENDOR_ID,
        subdevice:   PCI_ANY_ID,
        class:       0,
        class_mask:  0,
        driver_data: 0
    },
    {0, }
};

MODULE_DEVICE_TABLE(pci, tpropci_pci_tbl);

/*
**  tpropci_driver PCI driver information structure
*/
static struct pci_driver tpropci_driver =
{
    name:     TPRO_PCI_DRV_NAME,
    id_table: tpropci_pci_tbl,
    probe:    tpropci_init_one,
    remove:   __devexit_p(tpropci_remove_one),
    shutdown: tpropci_remove_one,
};

/*
** Driver setup structure
*/
static struct file_operations tpropci_fops =
{
    owner:   THIS_MODULE,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35)
    unlocked_ioctl:   tpropci_ioctl,
#else
    ioctl:   tpropci_ioctl,
#endif
    open:    tpropci_open,
    release: tpropci_close
};

static struct class         *tpropci_cs;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
static struct device        *tpropci_cd[TPRO_PCI_MAX_NUM];
#else
static struct class_device  *tpropci_cd[TPRO_PCI_MAX_NUM];
#endif
static struct cdev           tpropci_cdev;     /* Character dev struct.     */
static        dev_t          tpropci_devNum;   /* Driver's base Dev Number  */
static        tpropci_dev_t *tpropci_devp[TPRO_PCI_MAX_NUM];
static        char           NameStr[TPRO_PCI_MAX_NUM][32];


// Define the 16 register interface functions
extern FuncTable_t tpropci_16bit;

// Define the 32 register interface functions
extern FuncTable_t tpropci_32bit;

// Define the CPCI register interface functions
extern FuncTable_t tpropci_cpci;


/*******************************************************************************
**
** Function: tpropci_cleanup()
**
** Description: Routine called just before the driver is removed from memory.
**
** Parameters:
**     IN: None
**
**     RETURNS: None
**
*******************************************************************************/
static void tpropci_cleanup(void)
{
    /* Unregister the driver with the PCI subsystem */
    pci_unregister_driver(&tpropci_driver);

    class_unregister(tpropci_cs);

    cdev_del(&tpropci_cdev);
    unregister_chrdev_region(tpropci_devNum, TPRO_PCI_MAX_NUM);

    TPRO_PRINT("%s removed\n", tpropci_driver_string);

} /* End - tpropci_cleanup() */

/*******************************************************************************
**
** Function: tpropci_close()
**
** Description: Close a tpropci device
**
** Parameters:
**     IN: *inode - inode pointer
**         *filp  - file pointer
**
**     RETURNS: -EIO - Couldn't find matching tpropci device
**                 0 - Success
**
*******************************************************************************/
static int tpropci_close(struct inode *inode, struct file *filp)
{
    tpropci_user_t * user = NULL;
    tpropci_dev_t *  hw = NULL;
    int status = 0;

    /* Retrieve the user structure from the file pointer. */
    user = (tpropci_user_t *)(filp->private_data);
    filp = NULL;
    if (user == NULL)
    {
        status = -EFAULT;
	TPRO_LOG_ERR;
        goto fxnExit;
    }

    /* Retrieve the device pointer from the user structure. */
    hw = user->pDevice;
    if (hw == NULL)
    {
        status = -EFAULT;
	TPRO_LOG(TPRO_LOG_ERROR,
                 ("[J%lu]%s: Null device pointer in user struct %p!!!\n",
		 jiffies, __FUNCTION__, user));
    }

    #ifdef TPRO_CHECK
        /* Retrieve the device pointer from the inode and verify. */
        if (hw != tpropci_find_device(inode))
        {
            status = -EFAULT;
            TPRO_LOG(TPRO_LOG_ERROR,
                     ("[J%lu]%s: Device pointer mismatch"
                      " in user struct %p!!!\n",
                      jiffies, __FUNCTION__, user));
            goto fxnExit;
        }
    #endif

    /*
    ** Remove the user structure from the list and deallocate it.
    ** If this is the only user, shut down the device.
    */

    TPRO_SPIN_LOCK(&hw->userLock);

    #ifdef TPRO_CHECK
        /* Make sure the first and last user pointers are not NULL. */
        if ((hw->pFirstUser == NULL) || (hw->pLastUser == NULL))
        {
	    TPRO_SPIN_UNLOCK(&hw->userLock);
            status = -EFAULT;
            TPRO_LOG(TPRO_LOG_ERROR,
                     ("[J%lu]%s: hw->pFirstUser=%p, hw->pLastUser=%p in"
                      " device struct %p!!!\n",
                     jiffies, __FUNCTION__, hw->pFirstUser, hw->pLastUser,
                     user));
            goto fxnExit;
        }
    #endif

    if (user == hw->pFirstUser)
    {
        if (user == hw->pLastUser)
	{
	    /* This is the only user. */
            #ifdef TPRO_CHECK
                if ((user->pPrev != user) || (user->pNext != user))
		{
		    TPRO_LOG(TPRO_LOG_ERROR,
                             ("[J%lu]%s: Only user is %p but pPrev = %p"
                              " and pNext = %p",
                             jiffies, __FUNCTION__, user, user->pPrev,
                             user->pNext));
	        }
            #endif
            hw->pFirstUser = NULL;
            hw->pLastUser = NULL;
        }
        else
        {
            /* This is the first, but not the only, user. */
            user->pPrev->pNext = hw->pFirstUser = user->pNext;
            user->pNext->pPrev = user->pPrev;
        }
    }
    else if (user == hw->pLastUser)
    {
        /* This user was added last and is not the only user. */
        user->pNext->pPrev = hw->pLastUser = user->pPrev;
        user->pPrev->pNext = user->pNext;
    }
    else
    {
        /* Neither first nor last user.  Remove from the middle of the list. */
        user->pNext->pPrev = user->pPrev;
        user->pPrev->pNext = user->pNext;
    }

    TPRO_SPIN_UNLOCK(&hw->userLock);

    /* Free the user structure. */
    kfree(user);

    TPRO_LOG(TPRO_LOG_CLOSE,
             ("[J%lu]%s: Closed user struct %p\n",
              jiffies, __FUNCTION__, user));

 fxnExit:
    return status;

} /* End - tpropci_close() */

/*******************************************************************************
**
** Function: tpropci_find_device()
**
** Description: Determines which tpropci_dev_t stucture the inode is referring
**              to and return a pointer to it.  Returns NULL if no matching
**              tpropci_dev_t could be found.
**
** Parameters:
**     IN: *inode - info to match with a tpropci_dev_t sturcture
**
**     RETURNS: tpropci_dev_t* - NULL if none is found.
**
*******************************************************************************/
static tpropci_dev_t* tpropci_find_device(struct inode *inode)
{
    tpropci_dev_t *retDevp = NULL;
    unsigned int   idx;

    TPRO_LOG(TPRO_LOG_FIND_DEV,
             ("[J%lu]%s: Finding device, INODE=%p, DEVNUM=0x%08X\n",
              jiffies, __FUNCTION__, inode, inode->i_rdev));

    for (idx = 0; idx < TPRO_PCI_MAX_NUM; idx++)
    {
        if (tpropci_devp[idx] != NULL)
        {
            if (tpropci_devp[idx]->devNum == inode->i_rdev)
            {
                retDevp = tpropci_devp[idx];
            }
        }
    }

    return retDevp;

} /* End - tpropci_find_device() */

/*******************************************************************************
**
** Function: tpropci_init()
**
** Description: TPRO kernel module entry point
**
** Parameters:
**     IN: None
**
**     RETURNS: 0 on success, or negative error code.
**
*******************************************************************************/
static int tpropci_init(void)
{
    int rv;

    /* Print the driver ID strings. */
    TPRO_PRINT("%s - version %s\n%s\n",
               tpropci_driver_string,
               tpropci_driver_version,
               tpropci_copyright);

    /*
    ** Create class
    */
    tpropci_cs = class_create(THIS_MODULE, TPRO_PCI_DRV_NAME);

    if (IS_ERR(tpropci_cs))
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ERR "[J%lu]%s: Error creating sysfs class.\n",
                  jiffies, __FUNCTION__));
        return PTR_ERR(tpropci_cs);
    }

    /*
    ** Allocate character device region
    */
    rv = alloc_chrdev_region(&tpropci_devNum,
                             TPRO_PCI_BASE_MINOR,
                             TPRO_PCI_MAX_NUM,
                             TPRO_PCI_DRV_NAME);

    if (rv < 0)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: FAILED alloc_chrdev_region() err=%d\n",
		  jiffies, __FUNCTION__, rv));
        class_destroy(tpropci_cs); 
        return rv;
    }

    /*
    ** Initialize char device structure
    */
    cdev_init(&tpropci_cdev, &tpropci_fops);

    tpropci_cdev.owner = THIS_MODULE;
    tpropci_cdev.ops   = &tpropci_fops;

    /*
    ** Add the char device to the system
    */
    rv = cdev_add(&tpropci_cdev, tpropci_devNum, TPRO_PCI_MAX_NUM);

    if (rv)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ERR "[J%lu]%s: FAILED cdev_add()\n",
                 jiffies, __FUNCTION__));
        unregister_chrdev_region(tpropci_devNum, TPRO_PCI_MAX_NUM);
        class_destroy(tpropci_cs);
        return rv;
    }

    /*
    **  register the driver with the PCI subsytem
    */
    rv = pci_register_driver(&tpropci_driver);

    if (rv) {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ERR "[J%lu]%s: pci_register_driver returned %d\n",
                  jiffies, __FUNCTION__, rv));
    }

    return rv;

} /* End - tpropci_init() */

/*******************************************************************************
**
** Function: tpropci_init_one()
**
** Description: Initialize new device to be handled by this driver
**
** Parameters:
**     IN: *pdev - PCI device structure pointer
**         *ent  - PCI device ID pointer
**
**     RETURNS: 0 - Success, <0 - Failure
**
*******************************************************************************/
static int tpropci_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
    int          rv = 0;
    unsigned int idx;
    unsigned char initBoardResult;

    TPRO_LOG(TPRO_LOG_LOAD,
             ("[J%lu]%s\n",
              jiffies, __FUNCTION__));

    /* Find the next available device pointer... */
    for (idx = 0; idx < TPRO_PCI_MAX_NUM; idx++)
    {
        /* If the currently indexed pointer hasn't been allocated yet... */
        if (tpropci_devp[idx] == NULL)
        {
            /* Have found available device pointer */
            break;
        }
    }

    /* If all device pointers have been already allocated... */
    if (idx >= TPRO_PCI_MAX_NUM)
    {
        TPRO_LOG(TPRO_LOG_LOAD,
                 (KERN_ALERT "[J%lu]%s, devices already allocated\n",
                  jiffies, __FUNCTION__));
        return -ENOSPC;
    }

    /* Allocate device structure for new device */
    tpropci_devp[idx] = kmalloc(sizeof(tpropci_dev_t), GFP_KERNEL);

    /* If memory couldn't be allocated... */
    if (tpropci_devp[idx] == NULL)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: No memory\n", jiffies, __FUNCTION__));
        return -ENOMEM;
    }

    /*
    **  prepare device driver access context
    */
    tpropci_devp[idx]->intCnt     = 0;
    tpropci_devp[idx]->pFirstUser = NULL;
    tpropci_devp[idx]->pLastUser  = NULL;
    tpropci_devp[idx]->headIndex = 0;

    TPRO_MUTEX_INIT(&tpropci_devp[idx]->deviceLock);
    TPRO_SPIN_LOCK_INIT(&tpropci_devp[idx]->userLock);

    /*
    **  set PCI configuration register to enable PCI I/O and Memory access
    */
    rv = pci_enable_device(pdev);
    if (rv) {
        if (tpropci_devp[idx]) {
            kfree(tpropci_devp[idx]);
            tpropci_devp[idx] = NULL;
        }
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: Enable PCI device failed\n",
                 jiffies, __FUNCTION__));
        return -EIO;
    }

    /*
    **  get base address of device and amount of I/O space requested
    */
    if (pdev->device == TPRO_U66_DEVICE_ID)
    {
        tpropci_devp[idx]->ioAddr  = pci_resource_start(pdev, IO_BASE_ADDR_REG_32);
        tpropci_devp[idx]->ioSize  = pci_resource_len(pdev,   IO_BASE_ADDR_REG_32);
    }
    else if ((pdev->subsystem_device == TPRO_CPCI_SUBSYSTEM_DEVICE_ID) ||
             (pdev->subsystem_device == TSAT_CPCI_SUBSYSTEM_DEVICE_ID) ||
             (pdev->subsystem_device == TPRO_PMC_SUBSYSTEM_DEVICE_ID)  ||
             (pdev->subsystem_device == TSAT_PMC_SUBSYSTEM_DEVICE_ID))
    {
        unsigned long phys_addr    = pci_resource_start(pdev, MEMORY_BASE_ADDR_REG_CPCI);
        tpropci_devp[idx]->ioSize  = pci_resource_len(pdev,   MEMORY_BASE_ADDR_REG_CPCI);
        tpropci_devp[idx]->ioAddr  = (unsigned long) ioremap_nocache(phys_addr, tpropci_devp[idx]->ioSize);
    }
    else
    {
        tpropci_devp[idx]->ioAddr  = pci_resource_start(pdev, IO_BASE_ADDR_REG);
        tpropci_devp[idx]->ioSize  = pci_resource_len(pdev,   IO_BASE_ADDR_REG);
    }

//    tpropci_devp[idx]->device  = pdev->device;
    tpropci_devp[idx]->options = pdev->subsystem_device;
    tpropci_devp[idx]->irq     = pdev->irq;
    tpropci_devp[idx]->devNum  = MKDEV(MAJOR(tpropci_devNum), idx);

    sprintf(NameStr[idx], "tpropci%d", idx);

    /* Select the function set based on the board type. */
    if (pdev->device == TPRO_U66_DEVICE_ID)
    {
        TPRO_PRINT("TPro_PCI 32 bit version of board: %s\n", NameStr[idx]);
        tpropci_devp[idx]->pFunctionTable = &tpropci_32bit;
    }
    else if ((pdev->subsystem_device == TPRO_CPCI_SUBSYSTEM_DEVICE_ID) ||
             (pdev->subsystem_device == TSAT_CPCI_SUBSYSTEM_DEVICE_ID) ||
             (pdev->subsystem_device == TPRO_PMC_SUBSYSTEM_DEVICE_ID)  ||
             (pdev->subsystem_device == TSAT_PMC_SUBSYSTEM_DEVICE_ID))
    {
        TPRO_PRINT("TPro_CPCI 32 bit version of board: %s\n", NameStr[idx]);
        tpropci_devp[idx]->pFunctionTable = &tpropci_cpci;
    }
    else 
    {
        TPRO_PRINT("TPro_PCI 16 bit version of board: %s\n", NameStr[idx]);
        tpropci_devp[idx]->pFunctionTable = &tpropci_16bit;
    }

    /*
    ** Create class device
    */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
    tpropci_cd[idx] = class_device_create(tpropci_cs, NULL,
                                           MKDEV(MAJOR(tpropci_devNum), idx),
                                           NULL, NameStr[idx]);

#else
    tpropci_cd[idx] = device_create(tpropci_cs, NULL,
                                           MKDEV(MAJOR(tpropci_devNum), idx),
                                           NULL, NameStr[idx]);
#endif



    if (IS_ERR(tpropci_cd[idx]))
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: device creation failed\n",
                 jiffies, __FUNCTION__));
        return PTR_ERR(tpropci_cd[idx]);
    }

    sprintf(NameStr[idx], "tpropci%d", MINOR(tpropci_devp[idx]->devNum));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#define IRQF_DISABLED SA_INTERRUPT
#define IRQF_SHARED   SA_SHIRQ
#endif

    /* Map device IRQ */
    rv = request_irq(tpropci_devp[idx]->irq,
                     tpropci_isr, IRQF_DISABLED | IRQF_SHARED,
                     NameStr[idx],
                     tpropci_devp[idx]);

    if (rv != 0)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
	         ("[J%lu]%s: Could NOT map IRQ!!!\n",
                 jiffies, __FUNCTION__));
        return -EIO;
    }

    /* Initialize the board.  Used to be done in IOCTL. */
    initBoardResult =
        tpropci_devp[idx]->pFunctionTable->InitializeBoard(tpropci_devp[idx]);
    if (initBoardResult != 0) {
        TPRO_LOG(TPRO_LOG_ERROR,
	         ("[J%lu]%s: Board Initialization failed.\n",
                 jiffies, __FUNCTION__));
        return 1;
    }
    
    tpropci_devp[idx]->slotPosition = (unsigned char)idx;

    return 0;

} /* End - tpropci_init_one() */

/*******************************************************************************
**
** Function: tpropci_ioctl()
**
** Description: Called when an ioctl operation is requested.
**
** Parameters:
**     IN: *inode - inode struct for the device
**         *filp  - file struct for the device
**          cmd   - the ioctl command
**          arg   - typecast for the arg, can be a value/ptr
**
**     RETURNS: Negative on failure or the ioctl call return value (if any)
**
*******************************************************************************/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35)
static long tpropci_ioctl(struct file   *filp,
                         unsigned int  cmd,
                         unsigned long arg)
#else
static int tpropci_ioctl(struct inode  *inode,
                         struct file   *filp,
                         unsigned int  cmd,
                         unsigned long arg)
#endif
{
    tpropci_ioctl_cmd_t *cmdHnd     = (tpropci_ioctl_cmd_t*)arg;
    tpropci_dev_t       *hw         = NULL;
    tpropci_user_t      *user       = NULL;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35)
    long                  status     = 0;
#else
    int                  status     = 0;
#endif

    /* Validate the command. */
    if (_IOC_NR(cmd) >= IOCTL_TPRO_CMD_COUNT)
    {
        status = -EFAULT;
        TPRO_LOG(TPRO_LOG_ERROR,
		 ("[J%lu]%s: [CMD:0x%08X, USER:%p]: Invalid IOCTL command\n",
		 jiffies, __FUNCTION__, cmd, user));
        goto fxnExit;
    }

    /* Retrieve the user structure from the file pointer. */
    user = (tpropci_user_t *)(filp->private_data);
    if (user == NULL)
    {
        status = -EFAULT;
	TPRO_LOG_ERR;
        goto fxnExit;
    }

    TPRO_LOG(TPRO_LOG_IOCTL,
             ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]\n",
             jiffies, __FUNCTION__, cmd, user, user->pDevice));

    /* Retrieve the device pointer from the user structure. */
    hw = user->pDevice;
    if (hw == NULL)
    {
        status = -EFAULT;
        TPRO_LOG(TPRO_LOG_ERROR,
                 ("[J%lu]%s: [CMD:0x%08X, USER:%p]: Null device pointer"
                  " in user struct!!!\n",
		  jiffies, __FUNCTION__, cmd, user));
        goto fxnExit;
    }

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,35)
    #ifdef TPRO_CHECK
        /* Retrieve the device pointer from the inode and verify. */
        if (hw != tpropci_find_device(inode))
        {
 	    TPRO_LOG(TPRO_LOG_ERROR,
                     ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]"
                      ": Device pointer mismatch!!!\n",
                     jiffies, __FUNCTION__, cmd, user, user->pDevice));
        }
    #endif
#endif

    /* Verify access to the command object (read/write) */
    if (!access_ok(VERIFY_WRITE, (void*)cmdHnd, sizeof(tpropci_ioctl_cmd_t)))
    {
        status = -EFAULT;
        goto fxnExit;
    }

    /* Make sure the call is currently available to the user. */
    if ((cmd == IOCTL_TPRO_GET_TIME) || (cmd == IOCTL_TPRO_OPEN) ||
        (cmd == IOCTL_TPRO_SYNCH_STATUS) || (cmd == IOCTL_TPRO_GET_DRIVER) ||
        (cmd == IOCTL_TPRO_PEEK) || (cmd == IOCTL_TPRO_POKE))
    {
        /* These commands are available to everyone. */
    }
    else
    {
        int firstUser;

        /* These commands are available only to the first user. */
        TPRO_SPIN_LOCK(&hw->userLock);
        firstUser = (hw->pFirstUser == user) ? 1 : 0;
        TPRO_SPIN_UNLOCK(&hw->userLock);
	if (firstUser == 0)
        {
            status = -EFAULT;
            TPRO_LOG(TPRO_LOG_INFO_1,
                     ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]"
                      ": Multiuser conflict!!!\n",
		     jiffies, __FUNCTION__, cmd, user, user->pDevice));
            goto fxnExit;
	}
    }

    switch (cmd)
    {
    case IOCTL_TPRO_OPEN:
    {
        BoardObj board;

        /* verify access to board object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(BoardObj))) {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy object from user space */
        if (copy_from_user(&board, &cmdHnd->argVec[0], sizeof (BoardObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy board options to board object */
        board.options = hw->options;

        /* copy local object to user object */
        if (copy_to_user(&cmdHnd->argVec[0], &board, sizeof (BoardObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_ALTITUDE:
    {
        AltObj drvAlt;

        /* verify access to altitude object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(AltObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get altitude from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getAltitude(hw, &drvAlt);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvAlt, sizeof (AltObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_DATE:
    {
        DateObj drvDate;

        /* verify access to date object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(DateObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get date from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getDate(hw, &drvDate);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvDate, sizeof(DateObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_DRIVER:
    {
        char version[7];

        /* verify access to version string (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(version)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy version to local array */
        strcpy(version, VERSION);

        /* copy string to user space */
        if (copy_to_user(&cmdHnd->argVec[0], version, sizeof (version)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_FIRMWARE:
    {
        unsigned char firmware[5];

        /* verify access to firmware bytes (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(firmware)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get firmware from the tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getFirmware(hw, firmware);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy string to user space */
        if (copy_to_user(&cmdHnd->argVec[0], firmware, sizeof (firmware)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_LATITUDE:
    {
        LatObj drvLat;

        /* verify access to latitude object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(LatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get latitude from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getLatitude(hw, &drvLat);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvLat, sizeof (LatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_LONGITUDE:
    {
        LongObj drvLong;

        /* verify access to longitude object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (LongObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get longitude from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getLongitude(hw, &drvLong);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvLong, sizeof (LongObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_SAT_INFO:
    {
        SatObj drvSat;

        /* verify access to satellite info object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (SatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get satellite info from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getSatInfo(hw, &drvSat);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvSat, sizeof (SatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_TIME:
    {
        TimeObj drvTime;

        /* verify access to time object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (TimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get time from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getTime(hw, &drvTime);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvTime, sizeof (TimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_RESET_FIRMWARE:
    {
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->resetFirmware(hw);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_HEARTBEAT:
    {
        HeartObj drvHeart;

        /* copy object from user space */
        if (copy_from_user(&drvHeart, &cmdHnd->argVec [0], sizeof (HeartObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* set heartbeat on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setHeartbeat(hw, &drvHeart);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_MATCHTIME:
    {
        MatchObj drvMatch;

        /* copy object from user space */
        if (copy_from_user(&drvMatch, &cmdHnd->argVec[0], sizeof (MatchObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* set match time on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setMatchTime(hw, &drvMatch);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_PROP_DELAY_CORR:
    {
        int microseconds;

        /* If parameter is NULL, just return success. */
        if (cmdHnd == NULL)
	{
	    break;
	}

        /* get microsecond value from user */
        __get_user(microseconds, (int *) &cmdHnd->argVec[0]);

        /* set propagation delay correction value on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setPropDelayCorr(hw, &microseconds);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_TIME:
    {
        TimeObj drvTime;

        /* copy object from user space */
        if (copy_from_user(&drvTime, &cmdHnd->argVec [0], sizeof (TimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* set time on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setTime(hw, &drvTime);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_YEAR:
    {
        unsigned short year;

        /* get year value from user */
        __get_user(year, (unsigned short *) &cmdHnd->argVec[0]);

        /* set year to tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setYear(hw, &year);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SIM_EVENT:
    {
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->simEvent(hw);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SYNCH_CONTROL:
    {
        int ctrl;

        /* get control value from user */
        __get_user(ctrl, (int *) &cmdHnd->argVec[0]);

        /* set synchronization control value to tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->synchControl(hw, (unsigned char*)&ctrl);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SYNCH_STATUS:
    {
        unsigned char stat = 0xFF;

        /* verify access to synch status character (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (unsigned char)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get synchronization status from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->synchStatus(hw, &stat);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy status to user-space */
        if (copy_to_user(&cmdHnd->argVec[0], &stat, sizeof (unsigned char)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_WAIT_EVENT:
    {
        WaitObj drvWaitEvent;

        /* verify access to wait object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (WaitObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy object from user space */
        if (copy_from_user(&drvWaitEvent,
                           &cmdHnd->argVec[0],
                           sizeof (WaitObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, FIFO_IRQ_ENABLE);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* wait for event if necessary */
        if (hw->headIndex == hw->tailIndex)
        {
            /* clear flag condition */
            atomic_set (&hw->eventFlag, 0);

            /* wait on semaphore */
            status = os_waitPend(&hw->event_wait,
                                 drvWaitEvent.jiffies,
                                 &hw->eventFlag);

            /* return on timeout or waitaphore error */
            if (status == OS_WAIT_STATUS_TIMEOUT)
            {
                TPRO_MUTEX_LOCK(&hw->deviceLock);
                (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
                TPRO_MUTEX_UNLOCK(&hw->deviceLock);
                return 1;
            }

            if (status == OS_WAIT_STATUS_NG)
            {
                TPRO_MUTEX_LOCK(&hw->deviceLock);
                (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
                TPRO_MUTEX_UNLOCK(&hw->deviceLock);
                return -1;
            }
        }

        /* read event buffer into object */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
        (void)hw->pFunctionTable->getEvent(hw, &drvWaitEvent);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object back to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvWaitEvent, sizeof (WaitObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_WAIT_HEART:
    {
        int ticks;

        /* get microsecond value from user */
        __get_user(ticks, (int *) &cmdHnd->argVec[0]);

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, HEART_IRQ_ENABLE);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* clear flag condition */
        atomic_set(&hw->heartFlag, 0);

        /* wait on semaphore */
        status = os_waitPend(&hw->heart_wait, ticks, &hw->heartFlag);

        if (status == OS_WAIT_STATUS_TIMEOUT) status = 1;
        if (status == OS_WAIT_STATUS_NG) status = -1;

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_WAIT_MATCH:
    {
        int ticks;

        /* get microsecond value from user */
        __get_user(ticks, (int *) &cmdHnd->argVec[0]);

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, MATCH_IRQ_ENABLE);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* clear flag condition */
        atomic_set(&hw->matchFlag, 0);

        /* wait on semaphore */
        status = os_waitPend(&hw->match_wait, ticks, &hw->matchFlag);

        if (status == OS_WAIT_STATUS_TIMEOUT) status = 1;
        if (status == OS_WAIT_STATUS_NG) status = -1;

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_PEEK:
    {
        MemObj Reg;

        /* verify access to memory object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy object from user space */
        if (copy_from_user(&Reg, &cmdHnd->argVec [0], sizeof (MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get call the peek routine */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->peek(hw, &Reg);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object back to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &Reg, sizeof (MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_POKE:
    {
        MemObj Reg;

        /* copy object from user space */
        if (copy_from_user(&Reg, &cmdHnd->argVec [0], sizeof (MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* call the poke routine */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->poke(hw, &Reg);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_GET_NTP_TIME:
    {
        NtpTimeObj ntpTime;

        /* verify access to NTP time object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (NtpTimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get time from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getNtpTime(hw, &ntpTime);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* If subsystem device id is a TPRO family reference */
        if ((hw->options == TPRO_CPCI_SUBSYSTEM_DEVICE_ID) ||
            (hw->options == TPRO_PMC_SUBSYSTEM_DEVICE_ID))
        {
            memcpy(&ntpTime.refId, "IRIG", 4);
        }
        else if ((hw->options & 0xFF) < (TSAT_PCI & 0xFF))  
        {
            memcpy(&ntpTime.refId, "IRIG", 4);
        }
        else /* Else, TSAT reference */
        {
            memcpy(&ntpTime.refId, "GPS ", 4);
        }

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &ntpTime, sizeof(NtpTimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_FPGA:
    {
        unsigned char fpgaVer[6];

        /* verify access to firmware bytes (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(fpgaVer)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get firmware from the tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getFpgaVersion(hw, fpgaVer);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy string to user space */
        if (copy_to_user(&cmdHnd->argVec[0], fpgaVer, sizeof (fpgaVer)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_SET_OSCILLATOR:
    {
        unsigned char freq;

        /* get frequency value from user */
        __get_user(freq, (unsigned char *) &cmdHnd->argVec[0]);

        /* set frequency to tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setOsc(hw, &freq);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    default:
    {
	TPRO_LOG(TPRO_LOG_INFO_1,
             ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]"
              ": Unhandled command\n",
             jiffies, __FUNCTION__, cmd, user, user->pDevice));
        status = 1;
        goto fxnExit;
    }
    }

    fxnExit:
        return status;

} /* End - tpropci_ioctl() */

/*******************************************************************************
**
** Function:    tpropci_isr()
** Description: The tpropci driver ISR routine
**
** Parameters:
**     IN:  irq    - interrupt
**         *dev_id - tpropci device pointer
**         *regs   - processor's context
**
**     RETURNS: IRQ_HANDLED - IRQ was handled properly
**
*******************************************************************************/
static irqreturn_t tpropci_isr(ISR_ARGUMENT_LIST)
{
    tpropci_dev_t* hw = (tpropci_dev_t *) dev_id;


    return ((hw->pFunctionTable->isr(hw, &irq) == 0) ? IRQ_HANDLED : IRQ_NONE);

} /* End - tpropci_isr() */

/*******************************************************************************
**
** Function: tpropci_open()
**
** Description: Open the TPRO-PCI device.
**
** Parameters:
**     IN: *inode - inode struct for the device
**         *file  - file pointer to return TPRO-PCI object reference with
**
**     RETURNS: 0 on success, or negative error code.
**
*******************************************************************************/
static int tpropci_open(struct inode *inode, struct file *filp)
{
    tpropci_dev_t * hw;
    tpropci_user_t *user;

    hw = tpropci_find_device(inode);

    if (hw == NULL)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
		 ("[J%lu]%s: Couldn't find matching device!!!\n",
                 jiffies, __FUNCTION__));
        return -EIO;
    }

    /* Allocate user structure and store device pointer. */
    user = kmalloc (sizeof(tpropci_user_t), GFP_KERNEL);
    if (user == NULL)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
		 (KERN_ALERT "[J%lu]%s: No memory\n",
                 jiffies, __FUNCTION__));
        return -ENOMEM;
    }
    user->pDevice = hw;

    /* Lock the user list access.  If this is the first user, the lock must be maintained
     * around the device initialization as well. */
    TPRO_SPIN_LOCK(&hw->userLock);

    if (hw->pFirstUser != NULL)
    {
        /* Not the first user, so just insert the user structure into the linked list. */
        user->pNext = hw->pFirstUser;
        user->pPrev = hw->pLastUser;
        user->pNext->pPrev = user->pPrev->pNext = hw->pLastUser = user;
    }
    else
    {
        /* First user, so initialize the device. */
        hw->intCnt = 0;
        hw->eventCnt = 0;
        hw->headIndex = hw->tailIndex = 0;

        /* Since this is the first user, everything points to this user structure. */
        user->pNext = user->pPrev = hw->pFirstUser = hw->pLastUser = user;
    }

    /* Release access to the user list. */
    TPRO_SPIN_UNLOCK(&hw->userLock);

    TPRO_LOG(TPRO_LOG_OPEN,
	     ("[J%lu]%s: Adding user struct %p to device %p\n",
              jiffies, __FUNCTION__, user, hw));

    /* set private data */
    filp->private_data = user;

    return 0;

} /* End - tpropci_open */


/*******************************************************************************
**
** Function: tpropci_remove_one()
**
** Description: Removes a TPRO PCI device from the system
**
** Parameters:
**     IN: *pdev - tpropci device pointer
**
**     RETURNS: None
**
*******************************************************************************/
static void tpropci_remove_one(struct pci_dev *pdev)
{
    tpropci_dev_t       *hw = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
    struct device *cd = NULL;
#else
    struct class_device *cd = NULL;
#endif
    unsigned int         idx;

    for (idx = 0; idx < TPRO_PCI_MAX_NUM; idx++)
    {
        if (tpropci_devp[idx] != NULL)
        {
            if (tpropci_devp[idx]->irq == pdev->irq)
            {
                hw = tpropci_devp[idx];
                cd = tpropci_cd[idx];
                tpropci_devp[idx] = NULL;
                tpropci_cd[idx]   = NULL;
                break;
            }
        }
    }


    if (hw)
    {
        /* Disable interrupts on the device. */
        (void)hw->pFunctionTable->InitializeBoard(hw);

        /* Free the IRQ. */
        free_irq(hw->irq, hw);

        /* Destroy the spinlock and mutex. */
        TPRO_SPIN_LOCK_DESTROY(&hw->userLock);
        TPRO_MUTEX_DESTROY(&hw->deviceLock);

        // Set configuration register to disable PCI I/O & Memory Access
        pci_disable_device(pdev);

        /* Free the device object. */
        kfree(hw);
    }

    if (cd)
    {
        /* Destroy class device. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
        class_device_destroy(tpropci_cs,
                             MKDEV(MAJOR(tpropci_devNum), idx));

#else
        device_destroy(tpropci_cs,
                             MKDEV(MAJOR(tpropci_devNum), idx));
#endif

    }

} /* End - tpropci_remove_one() */


/*******************************************************************************
            Explicit Module Routine Declarations
*******************************************************************************/
module_init(tpropci_init);    /*-- init module ---------*/
module_exit(tpropci_cleanup); /*-- cleanup module ------*/
