/*******************************************************************************
**
**  Module  : tpro_func_cpci.c
**  Date    : 08/06/2012
**  Purpose : This driver provides an interface to the TPRO-PCI
**
**  Copyright(C) 2012 Spectracom Corporation. All Rights Reserved.
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**
*******************************************************************************/
#include <asm/delay.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/io.h>

#include "../h/ddtpro.h"
#include "tpropci.h"
#include "tpro_func.h"
#include "os_wait.h"

/*******************************************************************************
**          Definitions and Macros
*******************************************************************************/

/**
***  REGISTER DEFINITIONS
**/
#define TPRO_CPCI_STAT_REG       (0x00)
#define TPRO_CPCI_IRQ_ENB        (0x00)
            
#define TPRO_CPCI_CLK_UPPER      (0x04)
#define TPRO_CPCI_CLR_FLAG_M     (0x04)
            
#define TPRO_CPCI_CLK_LOWER      (0x08)
#define TPRO_CPCI_CLR_FLAG_HB    (0x08)
            
#define TPRO_CPCI_CLK_DATE       (0x0C)
#define TPRO_CPCI_CLR_FLAG_CMOV  (0x0C)
            
#define TPRO_CPCI_TTAG_STAT      (0x10)
#define TPRO_CPCI_SIM_TTAG       (0x10)
            
#define TPRO_CPCI_TTAG_UPPER     (0x14)
#define TPRO_CPCI_CLR_FLAG_SC    (0x14)
            
#define TPRO_CPCI_TTAG_LOWER     (0x18)
            
#define TPRO_CPCI_TTAG_DATE      (0x1C)
#define TPRO_CPCI_CLR_FLAG_TT    (0x1C)
            
#define TPRO_CPCI_CMD_REG_0      (0x20)
#define TPRO_CPCI_CMD_REG_1      (0x24)
#define TPRO_CPCI_CMD_REG_2      (0x28)
#define TPRO_CPCI_CMD_REG_3      (0x2C)
            
#define TPRO_CPCI_RESP_REG_0     (0x30)
#define TPRO_CPCI_RESP_REG_1     (0x34)
#define TPRO_CPCI_RESP_REG_2     (0x38)
#define TPRO_CPCI_RESP_REG_3     (0x3C)
            
#define TPRO_CPCI_RESET_REG      (0x40)

/**
***  REGISTER BIT DEFINITIONS
**/
#define TPRO_CPCI_M_EN           (0x1 << 8)
#define TPRO_CPCI_HB_EN          (0x1 << 9)
#define TPRO_CPCI_TT_EN          (0x1 << 10)

/*******************************************************************************
**          Public function pointer table
*******************************************************************************/
FuncTable_t tpropci_cpci =
{
    tproGetAltitudeCPCI,
    tproGetDateCPCI,
    tproGetEventCPCI,
    tproGetFirmwareCPCI,
    tproInitializeBoardCPCI,
    tproGetLatitudeCPCI,
    tproGetLongitudeCPCI,
    tproGetSatInfoCPCI,
    tproGetTimeCPCI,
    tproGetNtpTimeCPCI,
    tproResetFirmwareCPCI,
    tproSetHeartbeatCPCI,
    tproSetMatchTimeCPCI,
    tproSetPropDelayCorrCPCI,
    tproSetTimeCPCI,
    tproSetYearCPCI,
    tproSimEventCPCI,
    tproSynchControlCPCI,
    tproSynchStatusCPCI,
    tproPeekCPCI,
    tproPokeCPCI,
    tproGetFpgaVersionCPCI,
    tproToggleInterruptCPCI,
    tproSetOscCPCI,
    tproISRCPCI,


    TPRO_CPCI_STAT_REG,
    TPRO_CPCI_IRQ_ENB,
    TPRO_CPCI_IRQ_ENB
};


/*******************************************************************************
**          Private Funtion Prototypes
*******************************************************************************/
unsigned short getResponseCPCI (unsigned long base, char *str);
void           sendCmdCPCI     (unsigned long base, unsigned int *cmdArray);


/***************************************************************************
**          Private Function Definitions
***************************************************************************/

/*******************************************************************************
** Function: getResponseCPCI
*******************************************************************************/
unsigned short getResponseCPCI(unsigned long base, char *str)
{
    unsigned int response[4];


    // read response registers
    response[0] = PEEK_REG_CPCI(base, TPRO_CPCI_RESP_REG_0);
    response[1] = PEEK_REG_CPCI(base, TPRO_CPCI_RESP_REG_1);
    response[2] = PEEK_REG_CPCI(base, TPRO_CPCI_RESP_REG_2);
    response[3] = PEEK_REG_CPCI(base, TPRO_CPCI_RESP_REG_3);

/*
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: RESPONSE(%d) = 0x%08X\n", 0, response[0]));
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: RESPONSE(%d) = 0x%08X\n", 1, response[1]));
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: RESPONSE(%d) = 0x%08X\n", 2, response[2]));
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: RESPONSE(%d) = 0x%08X\n", 3, response[3]));
*/
    // fill-in string
    *str++ = (response[0] & 0xFF)       >> 0;
    *str++ = (response[0] & 0xFF00)     >> 8;
    *str++ = (response[0] & 0xFF0000)   >> 16;
    *str++ = (response[0] & 0xFF000000) >> 24;

    *str++ = (response[1] & 0xFF)       >> 0;
    *str++ = (response[1] & 0xFF00)     >> 8;
    *str++ = (response[1] & 0xFF0000)   >> 16;
    *str++ = (response[1] & 0xFF000000) >> 24;

    *str++ = (response[2] & 0xFF)       >> 0;
    *str++ = (response[2] & 0xFF00)     >> 8;
    *str++ = (response[2] & 0xFF0000)   >> 16;
    *str   = (response[2] & 0xFF000000) >> 24;


   // return response id
    return ((short)(response[3] & 0xFFFF));
}


/*******************************************************************************
** Function: sendCmdCPCI
*******************************************************************************/
void sendCmdCPCI(unsigned long base, unsigned int *cmdArray)
{
 
    // send command sequence
    POKE_REG_CPCI(base, TPRO_CPCI_CMD_REG_0, cmdArray[0]); 
    POKE_REG_CPCI(base, TPRO_CPCI_CMD_REG_1, cmdArray[1]);
    POKE_REG_CPCI(base, TPRO_CPCI_CMD_REG_2, cmdArray[2]);
    POKE_REG_CPCI(base, TPRO_CPCI_CMD_REG_3, cmdArray[3]);

 
    while (!(PEEK_REG_CPCI (base, TPRO_CPCI_STAT_REG) & 0x40));


/*
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: COMMAND(%d) = 0x%08X\n", 0, cmdArray[0]));
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: COMMAND(%d) = 0x%08X\n", 1, cmdArray[1]));
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: COMMAND(%d) = 0x%08X\n", 2, cmdArray[2]));
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROPCI: COMMAND(%d) = 0x%08X\n", 3, cmdArray[3]));
*/


} // End - sendCmdCPCI() */


/*******************************************************************************
**          Public Function Definitions
*******************************************************************************/

/*******************************************************************************
**
** Function:    tproToggleInterruptCPCI()
** Description: toggle an interrupt flag
**
** Parameters:
**     IN: *hw   - Handle
**         mask  -  interrupt mask to write to ICR register
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproToggleInterruptCPCI (tpropci_dev_t *hw, unsigned char mask)
{
    unsigned int newMask = 0x4000; //set TimeTag input on

    // Convert to cPCI interrupt masking:
    //  Match
    if (mask & MATCH_IRQ_ENABLE)
    {
        newMask |= TPRO_CPCI_M_EN;
    }
    //  Heartbeat
    if (mask & HEART_IRQ_ENABLE)
    {
        newMask |= TPRO_CPCI_HB_EN;
    }
    //  Time Tag
    if (mask & FIFO_IRQ_ENABLE)
    {
        newMask |= TPRO_CPCI_TT_EN;
    }

    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_IRQ_ENB, newMask);

    return (0);
}


/*******************************************************************************
**
** Function:    tproGetAltitudeCPCI()
** Description: Get altitude from the TPRO device
**
** Parameters:
**     IN: *hw   - Handle
**         *Altp - Pointer the Altitude result
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetAltitudeCPCI(tpropci_dev_t *hw, AltObj *Altp)
{
    unsigned int command [4];
    char         response[13];
    char         *altString;



    // send retrieve altitude command
    command[0] = 0x00;
    command[1] = 0x00;
    command[2] = 0x00;
    command[3] = 0x70;

    sendCmdCPCI(hw->ioAddr, command);

    Altp->meters = 0; // pre clear altitude to zero

    // get response from card and verify its the altitude response
    if (getResponseCPCI(hw->ioAddr, response) != 0x70)
    {
        return (-1);
    }

    if ((altString = strchr(response,'M')) !=NULL) //if no sat's seen
    {  
        // parse response, convert to int and * 10 to scale correctly
        sscanf (response,"%d", &Altp->meters);
        Altp->meters *= 10;
 
    }

    return (0);

} /* End - tproGetAltitudeCPCI() */


/*******************************************************************************
**
** Function:    tproGetDateCPCI()
** Description: Gets the date in Gregorian format from the TPRO device.
**
** Parameters:
**     IN: *hw    - Handle
**         *Datep - Pointer to date result
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetDateCPCI(tpropci_dev_t *hw, DateObj *Datep)
{
    unsigned int val;


    // read status register -- this latches the instantaneous clock time and date
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_STAT_REG);

    // dummy reads for time
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_UPPER);
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_LOWER);

    // read date register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_DATE);

    // parse year
    Datep->year   = (unsigned short)(((val & 0x0000F000) >> 12) * 1000);
    Datep->year  += (unsigned short)(((val & 0x00000F00) >>  8) * 100);
    Datep->year  += (unsigned short)(((val & 0x000000F0) >>  4) * 10);  
    Datep->year  += (unsigned short)(((val & 0x0000000F) >>  0) * 1);  

    // parse day
    Datep->day    =  (unsigned char)(((val & 0x00F00000) >> 20) * 10);
    Datep->day   +=  (unsigned char)( (val & 0x000F0000) >> 16);

    // parse month
    Datep->month  =  (unsigned char)(((val & 0xF0000000) >> 28) * 10);
    Datep->month +=  (unsigned char)( (val & 0x0F000000) >> 24);

    return (0);

} /* End - tproGetDateCPCI() */


/*******************************************************************************
**
** Function:    tproGetEventCPCI()
** Description: Reads the time-tagged event information.
**
** Parameters:
**     IN: *hw    - Handle
**         *Waitp - Pointer to event information
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetEventCPCI(tpropci_dev_t *hw, WaitObj *Waitp)
{
    unsigned short *ptr = (unsigned short *)&hw->fifoVector[hw->tailIndex];

    /* parse days */
    Waitp->days = (unsigned short)((LO_NIBBLE     (ptr[2]) * 100) +
                                   (MID_LO_NIBBLE (ptr[3]) * 10)  +
                                   (LO_NIBBLE     (ptr[3]) * 1) );

    /* parse hours */
    Waitp->hours = (unsigned char)((MID_LO_NIBBLE (ptr[4]) * 10) +
                                   (LO_NIBBLE     (ptr[4]) * 1) );

    /* parse minutes */
    Waitp->minutes = (unsigned char)((MID_LO_NIBBLE (ptr[5]) * 10) +
                                     (LO_NIBBLE     (ptr[5]) * 1) );

    /* parse seconds */
    Waitp->seconds = (unsigned int)((LO_NIBBLE     (ptr[9]) * 1) +
                                    (MID_LO_NIBBLE (ptr[9]) * 10) +
                                    (LO_NIBBLE     (ptr[8]) * 100) +
                                    (MID_LO_NIBBLE (ptr[8]) * 1000) +
                                    (LO_NIBBLE     (ptr[7]) * 10000) +
                                    (MID_LO_NIBBLE (ptr[7]) * 100000) +
                                    (LO_NIBBLE     (ptr[6]) * 1000000) +
                                    (MID_LO_NIBBLE (ptr[6]) * 10000000) );

    /* increment tail index */
    hw->tailIndex = ((hw->tailIndex + 10) % (sizeof(hw->fifoVector)/sizeof(hw->fifoVector[0])));

    // read upper clock register
    PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_UPPER);

    // read lower clock register
    PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_LOWER);

    // read date
    PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_DATE);

    return (0);

} /* End - tproGetEventCPCI() */


/*******************************************************************************
**
** Function:    tproGetFirmwareCPCI()
** Description: Get Firmware ID from TPRO device.
**
** Parameters:
**     IN: *hw   - Handle
**         *firm - Pointer to firmware ID
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetFirmwareCPCI(tpropci_dev_t *hw, unsigned char *firm)
{
    unsigned int response[4];
    unsigned int command [4];


    // send retrieve firmware command
    command[0] = 0x00;
    command[1] = 0x00;
    command[2] = 0x00;
    command[3] = 0xEC;

    sendCmdCPCI(hw->ioAddr, command);
   

    // read response registers
    response[0] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_0);
    response[1] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_1);
    response[2] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_2);
    response[3] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_3);

    // verify response
    if ((response[3] & 0xFFFF) != 0x00EC)
    {
        return (-1);
    }

    // copy firmware version
    *firm++ = (response[2] & 0x00FF0000) >> 16;
    *firm++ = (response[2] & 0x0000FF00) >> 8;
    *firm++ = (response[2] & 0x000000FF) >> 0;
    *firm++ = (response[2] & 0x00000000) ;

    return (0);

} /* End - tproGetFirmwareCPCI() */


/*******************************************************************************
**
** Function:    tproGetLatitudeCPCI()
** Description: Get Latitude from TPRO device.
**
** Parameters:
**     IN: *hw   - Handle
**         *Latp - Pointer to latitude result
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetLatitudeCPCI(tpropci_dev_t *hw, LatObj *Latp)
{
    unsigned int command [4];
    char         response[13];


    // send retrieve latitude command
    command[0] = 0x00;
    command[1] = 0x00;
    command[2] = 0x00;
    command[3] = 0x72;

    sendCmdCPCI(hw->ioAddr, command);

    // preclear latitude
    Latp->degrees = 0;
    Latp->minutes = 0;


    // get response from card and verify its the latitude response
    if ((getResponseCPCI(hw->ioAddr, response)) != 0x72 )
    {
        return (-1);
    }

    if ((strchr(response,'.')) !=NULL) //if no sat's seen
    {  
        // parse degrees
        Latp->degrees  = (unsigned short)((CHAR_TO_DEC(response[0]) * 10) +
                                          (CHAR_TO_DEC(response[1]) *  1));
        // if southern hem
        if ((strchr(response,'S')) != NULL)
        {     
            Latp->degrees += 180 ;
        }
        // parse minutes
        Latp->minutes  =   (unsigned int)((CHAR_TO_DEC(response[2]) * 100000) +
                                          (CHAR_TO_DEC(response[3]) *  10000) +
                                          (CHAR_TO_DEC(response[5]) *   1000) +
                                          (CHAR_TO_DEC(response[6]) *    100) +
                                          (CHAR_TO_DEC(response[7]) *     10) );

    }


    return (0);

} /* End - tproGetLatitudeCPCI() */


/*******************************************************************************
**
** Function:    tproGetLongitudeCPCI()
** Description: Get Longitude from TPRO device.
**
** Parameters:
**     IN: *hw   - Handle
**         *Latp - Pointer to longitude result
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetLongitudeCPCI(tpropci_dev_t *hw, LatObj *Longp)
{
    unsigned int command [4];
    char         response[13];


    // send retrieve longitude command
    command[0] = 0x00;
    command[1] = 0x00;
    command[2] = 0x00;
    command[3] = 0x71;

    sendCmdCPCI(hw->ioAddr, command);

    // preclear longitude
    Longp->degrees = 0;
    Longp->minutes = 0;


    // get response from card and verify its the longitude response
    if ((getResponseCPCI(hw->ioAddr, response)) != 0x71)
    {
        return (-1);
    }


    if ((strchr(response,'.')) !=NULL) //if no sat's seen
    {  

        // parse degrees
        Longp->degrees  = (unsigned short)((CHAR_TO_DEC(response[0]) * 100) +
                                           (CHAR_TO_DEC(response[1]) *  10) +
                                           (CHAR_TO_DEC(response[2]) *   1));
        // if southern hem
        if ((strchr(response,'W')) != NULL)
        {     
            Longp->degrees += 180 ;
        }

        /**
        ***  parse minutes
        **/

        Longp->minutes  =   (unsigned int)((CHAR_TO_DEC(response[3]) * 100000) +
                                       (CHAR_TO_DEC(response[4]) *  10000) +
                                       (CHAR_TO_DEC(response[6]) *   1000) +
                                       (CHAR_TO_DEC(response[7]) *    100) +
                                       (CHAR_TO_DEC(response[8]) *     10));
    }


    return (0);

} /* End - tproGetLongitutdeCPCI() */


/*******************************************************************************
**
** Function:    tproInitializeBoardCPCI()
** Description: Initializes TPRO registers upon board creation.
**
** Parameters:
**     IN: *hw   - Handle
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproInitializeBoardCPCI(tpropci_dev_t *hw)
{
    /*
    **  initialize semaphores
    */
    /* heartbeat */
    os_waitInit(&hw->heart_wait);
    atomic_set(&hw->heartFlag, 0);

    /* match time */
    os_waitInit(&hw->match_wait);
    atomic_set(&hw->matchFlag, 0);

    /* external time tagged events */
    os_waitInit(&hw->event_wait);
    atomic_set(&hw->eventFlag, 0);

    // disable interrupts and enable timetag
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_IRQ_ENB, 0x4000); //set TimeTag input on
    PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_DATE);

    // clear interrupt flags
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_HB, 0x0);
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_M,  0x0);
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_TT, 0x0);
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_CMOV, 0x0);

    return (0);

} /* End - tproInitializeBoardCPCI() */


/*******************************************************************************
**
** Function:    tproGetSatInfoCPCI()
** Description: Get satellite information from TPRO device.
**
** Parameters:
**     IN: *hw   - Handle
**         *Satp - Pointer to satellite
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetSatInfoCPCI(tpropci_dev_t *hw, SatObj *Satp)
{
    unsigned int command [4];
    char         response[13];
    char         *satString;
    unsigned int satsTracked = 0;


    // send retrieve satellite command
    command[0] = 0x00;
    command[1] = 0x00;
    command[2] = 0x00;
    command[3] = 0x70;

    sendCmdCPCI(hw->ioAddr, command);


    Satp->satsTracked = 0; //pre clear #sat to zero

    // get response from card and verify its the sat info response
    if (getResponseCPCI(hw->ioAddr, response) != 0x70)
    {
      return (-1);
    }

    // find comma delimited altitude
    if (satString = strrchr (response, ','), satString != NULL)
    {
      // parse satellite tracked information
      sscanf (satString, ",M%d", &satsTracked);
      Satp->satsTracked = (unsigned char)satsTracked;
    }

    return (0);

} /* End - tproGetSatInfoCPCI() */


/*******************************************************************************
**
** Function:    tproGetTimeCPCI()
** Description: Get Time information from the TPRO device.
**
** Parameters:
**     IN: *hw    - Handle
**         *Timep - Pointer to time information
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetTimeCPCI(tpropci_dev_t *hw, TimeObj *Timep)
{
    unsigned int val;


    // read status register -- this latches the instantaneous clock time and
    // date
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_STAT_REG);

    // setup status flag info
    Timep->flags =  (unsigned short)(((val & 0x00000002) <<  1 ) +
                                     ((val & 0x00000001) <<  1 ));
    // indicate no flags info
//    Timep->flags = 0xffff;

    // read upper register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_UPPER);

    // parse days
    Timep->days       = (unsigned short)((((val & 0x0F000000) >> 24) * 100) +
                                         (((val & 0x00F00000) >> 20) *  10) +
                                         (((val & 0x000F0000) >> 16) *   1));

    // parse hours
    Timep->hours      =  (unsigned char)((((val & 0x0000F000) >> 12) * 10) +
                                         (((val & 0x00000F00) >>  8) *  1));

    // parse minutes
    Timep->minutes    =  (unsigned char)((((val & 0x000000F0) >>  4) * 10) +
                                         (((val & 0x0000000F) >>  0) *  1));

    // read lower clock register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_LOWER);

    // parse seconds
    Timep->secsDouble =   (unsigned int)((((val & 0xF0000000) >> 28) * 10000000) +
                                         (((val & 0x0F000000) >> 24) *  1000000) +
                                         (((val & 0x00F00000) >> 20) *   100000) +
                                         (((val & 0x000F0000) >> 16) *    10000) +
                                         (((val & 0x0000F000) >> 12) *     1000) +
                                         (((val & 0x00000F00) >>  8) *      100) +
                                         (((val & 0x000000F0) >>  4) *       10) +
                                         (((val & 0x0000000F) >>  0) *        1));

    // read date register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_DATE);

    // parse year
    Timep->year = (unsigned short)((((val & 0x0000F000) >> 12) *     1000) +
                                   (((val & 0x00000F00) >>  8) *      100) +
                                   (((val & 0x000000F0) >>  4) *       10) +  
                                   (((val & 0x0000000F) >>  0) *        1));  



    return (0);

} /* End - tproGetTimeCPCI() */

/*******************************************************************************
**
** Function:    tproGetNtpTimeCPCI()
** Description: Get Time information from the TPRO device and from the linux
**              kernel.
**
** Parameters:
**     IN: *hw    - Handle
**         *Timep - Pointer to time information
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetNtpTimeCPCI(tpropci_dev_t *hw, NtpTimeObj *Timep)
{
    unsigned int  val;


    /* get the time of day (kernel-style). */
    do_gettimeofday(&(Timep->tv));

    // read status register -- this latches the instantaneous clock time and
    // date
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_STAT_REG);

    // read upper register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_UPPER);

    // parse days
    Timep->timeObj.days       = (unsigned short)((((val & 0x0F000000) >> 24) * 100) +
                                                 (((val & 0x00F00000) >> 20) *  10) +
                                                 (((val & 0x000F0000) >> 16) *   1));

    // parse hours
    Timep->timeObj.hours      =  (unsigned char)((((val & 0x0000F000) >> 12) * 10) +
                                                 (((val & 0x00000F00) >>  8) *  1));

    // parse minutes
    Timep->timeObj.minutes    =  (unsigned char)((((val & 0x000000F0) >>  4) * 10) +
                                                 (((val & 0x0000000F) >>  0) *  1));

    // read lower clock register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_LOWER);

    // parse seconds
    Timep->timeObj.secsDouble =   (unsigned int)((((val & 0xF0000000) >> 28) * 10000000) +
                                                 (((val & 0x0F000000) >> 24) *  1000000) +
                                                 (((val & 0x00F00000) >> 20) *   100000) +
                                                 (((val & 0x000F0000) >> 16) *    10000) +
                                                 (((val & 0x0000F000) >> 12) *     1000) +
                                                 (((val & 0x00000F00) >>  8) *      100) +
                                                 (((val & 0x000000F0) >>  4) *       10) +
                                                 (((val & 0x0000000F) >>  0) *        1));

    // read date register
    val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLK_DATE);

    // parse year
    Timep->timeObj.year       = (unsigned short)((((val & 0x0000F000) >> 12) *     1000) +
                                                 (((val & 0x00000F00) >>  8) *      100) +
                                                 (((val & 0x000000F0) >>  4) *       10) +  
                                                 (((val & 0x0000000F) >>  0) *        1));  

    return (0);

} /* End - tproGetNtpTimeCPCI() */


/*******************************************************************************
**
** Function:    tproGetFpgaVersionCPCI()
** Description: get the FPGA version from the board.
**
** Parameters:
**     IN: *hw    - Handle
**         *fpagVer - Pointer to the char array.
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproGetFpgaVersionCPCI(tpropci_dev_t *hw, unsigned char *fpgaVer)
{
    unsigned int response[4];
    unsigned int command [4];


    // send retrieve firmware command
    command[0] = 0x00;
    command[1] = 0x00;
    command[2] = 0x00;
    command[3] = 0xEC;

    sendCmdCPCI(hw->ioAddr, command);

    // read response registers
    response[0] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_0);
    response[1] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_1);
    response[2] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_2);
    response[3] = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_3);

    // verify response
    if ((response[3] & 0xFFFF) != 0x00EC)
    {
        return (1);
    }

    *fpgaVer++= (response[0] & 0x00FF0000) >> 16;
    *fpgaVer++= (response[0] & 0x0000FF00) >>  8;
    *fpgaVer++= (response[0] & 0x000000FF) >>  0;


    return (0);

} /* End - tproGetFpgaVersionCPCI() */


/*******************************************************************************
**
** Function:    tproResetFirmwareCPCI()
** Description: Reset Firmware on the TPRO device.
**
** Parameters:
**     IN: *hw - Handle
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproResetFirmwareCPCI(tpropci_dev_t *hw)
{
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESET_REG, 0xFFFFFFFF);

    return (0);

} /* End - tproResetfirmwareCPCI() */


/*******************************************************************************
**
** Function:    tproSetHeartbeatCPCI()
** Description: Configures the heartbeat output on the TPRO device.
**
** Parameters:
**     IN: *hw     - Handle
**         *Heartp - Pointer to heartbeat information
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSetHeartbeatCPCI(tpropci_dev_t *hw, HeartObj *Heartp)
{
    unsigned int command[4] = {0};
    unsigned int control;


    // build control word (defaults to normal/enabled)
    control = 0x4 | (Heartp->outputType);  // outputType holds clock select

    // send heartbeat commands
    command[0] = Heartp->divNumber;
    command[1] = control;
    command[3] = 0x40;

    sendCmdCPCI(hw->ioAddr, command);

    return (0);

} /* End - tproSetHeartbeatCPCI() */


/*******************************************************************************
**
** Function:    tproSetMatchTimeCPCI()
** Description: Configures the match start/stop time feature.
**
** Parameters:
**     IN: *hw     - Handle
**         *Matchp - Pointer to start/stop information
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSetMatchTimeCPCI(tpropci_dev_t *hw, MatchObj *Matchp)
{
    unsigned int   command[4];
    char           timeStr[10];
    char          *timePtr;
    unsigned char  k;


    // initialize command registers
    command[0] = 0;
    command[1] = 0;
    command[2] = 0;
    command[3] = 0;

    // clear existing match condition if not already satisfied
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_M, 0xFF);

    // set days
    sprintf(timeStr, "%03d", Matchp->days);
    timePtr = timeStr;

    for (k=0; k < 3; k++) 
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (24 - k * 4));
    }

    // set hours
    sprintf(timeStr, "%02d", Matchp->hours);
    timePtr = timeStr;

    for (k=0; k < 2; k++) 
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (12 - k * 4));
    }

    // set minutes
    sprintf(timeStr, "%02d", Matchp->minutes);
    timePtr = timeStr;

    for (k=0; k < 2; k++)
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (4 - k * 4));
    }

    // set seconds
    sprintf(timeStr, "%08d", Matchp->seconds);
    timePtr = timeStr;

    for (k=0; k < 8; k++) 
    {
        command[1] += (CHAR_TO_DEC(*timePtr++) << (28 - k * 4));
    }

    // set match type
    command[3] = (Matchp->matchType) ? 0x30 : 0x20;

    // send command
    sendCmdCPCI(hw->ioAddr, command);

#if 0
    // verify response
    if (PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_3) == 0x20)
    {
        return (1);
    }
#endif

    return (0);

} /* End - tproSetMatchTimeCPCI() */


/*******************************************************************************
**
** Function:    tproSetPropDelayCorrCPCI()
** Description: Sets the propagation delay correction factor.
**
** Parameters:
**     IN: *hw           - Handle
**         *microseconds - Pointer to propagation delay correction
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSetPropDelayCorrCPCI(tpropci_dev_t *hw, int *microseconds)
{
    unsigned int   command[4];
    char           timeStr[10];
    char          *timePtr;
    unsigned char  k;


    // initialize command registers
    command[0] = 0;
    command[1] = 0;
    command[2] = 0;
    command[3] = 0x60;

    // build offset commands
    sprintf(timeStr, "%04d", *microseconds);
    timePtr = timeStr + 1;  // skip over +/- sign

    for (k=0; k < 3; k++) 
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (8 - k * 4));
    }
    command[0] |= (*microseconds < 0) ? 0 : 0x1000;

    // send commands
    sendCmdCPCI(hw->ioAddr, command);

    return (0);

} /* End - tproSetPropDelayCorrCPCI() */


/*******************************************************************************
**
** Function:    tproSetTimeCPCI()
** Description: Sets the TPRO device's IRIG-B generator to a given time.
**
** Parameters:
**     IN: *hw    - Handle
**         *Timep - Pointer to time value
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSetTimeCPCI(tpropci_dev_t *hw, TimeObj *Timep)
{
    unsigned int   command[4];
    char           timeStr[10];
    char          *timePtr;
    unsigned char  k;


    // initialize command registers
    command[0] = 0;
    command[1] = 0;
    command[2] = 0;
    command[3] = 0x10;

    // set days
    sprintf(timeStr, "%03d", Timep->days);
    timePtr = timeStr;

    for (k=0; k < 3; k++) 
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (24 - k * 4));
    }

    // set hours
    sprintf(timeStr, "%02d", Timep->hours);
    timePtr = timeStr;

    for (k=0; k < 2; k++) 
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (12 - k * 4));
    }

    // set minutes
    sprintf(timeStr, "%02d", Timep->minutes);
    timePtr = timeStr;

    for (k=0; k < 2; k++)
    {
        command[0] += (CHAR_TO_DEC(*timePtr++) << (4 - k * 4));
    }

    // set seconds
    sprintf(timeStr, "%02d", Timep->seconds);
    timePtr = timeStr;

    for (k=0; k < 2; k++) 
    {
        command[1] += (CHAR_TO_DEC(*timePtr++) << (28 - k * 4));
    }

    // set year
    sprintf(timeStr, "%04d", Timep->year);
    timePtr = timeStr;

    for (k=0; k < 4; k++) 
    {
        command[2] += (CHAR_TO_DEC(*timePtr++) << (12 - k * 4));
    }

    // jam in year value to ensure valid gregorian date reading
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CMD_REG_2, command[2]);

    // send command
    sendCmdCPCI(hw->ioAddr, command);

    return (0);

} /* End - tproSetTimeCPCI() */


/*******************************************************************************
**
** Function:    tproSetYearCPCI()
** Description: Sets the year on the TPRO device.
**
** Parameters:
**     IN: *hw   - Handle
**         *year - Pointer to time value
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSetYearCPCI(tpropci_dev_t *hw, unsigned short *year)
{
    unsigned int  command[4];
    char          yearStr[10];
    unsigned char k;


    // initialize command registers
    command[0] = 0;
    command[1] = 0;
    command[2] = 0;
    command[3] = 0;

    // set year
    sprintf(yearStr, "%04d", *year);

    for (k=0; k < 4; k++) 
    {
        command[2] += (CHAR_TO_DEC(yearStr[k]) << (12 - k * 4));
    }
    command[3] = 0x15;

    sendCmdCPCI(hw->ioAddr, command);

    return (0);

} /* End - tproSetYearCPCI() */


/*******************************************************************************
**
** Function:    tproSimEventCPCI()
** Description: Routine will simulate external time tag event.
**
** Parameters:
**     IN: *hw   - Handle
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSimEventCPCI(tpropci_dev_t *hw)
{
    // simulate event
    TPRO_LOG(TPRO_LOG_INFO_1, ("TPROCPCI: Simulating Event!\n"));
    POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_SIM_TTAG, 0xFFFFFFFF);

    return (0);

} /* End - tproSimEventCPCI() */


/*******************************************************************************
**
** Function:    tproSynchControlCPCI()
** Description: Routine determines whether to synchronize the on-board generator
**              to the TPRO's input or to disable synchronization.
**
** Parameters:
**     IN: *hw   - Handle
**         *enbp - (1) to enable synchronization
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSynchControlCPCI(tpropci_dev_t *hw, unsigned char *enbp)
{
    unsigned int  command[4];
    unsigned char enable = *enbp;


    // initialize command registers
    command[0] = 0;
    command[1] = 0;
    command[2] = 0;

    // enable
    if (enable == 1) 
    {
        *enbp      = 0xEE;
        command[3] = 0xC1;
    }
    // disable
    else 
    {
        *enbp      = 0xFF;
        command[3] = 0xC0;
    }

    sendCmdCPCI(hw->ioAddr, command);

    return (0);

} /* End - tproSynchControlCPCI() */


/*******************************************************************************
**
** Function:    tproSynchStatusCPCI()
** Description: Routine returns the state of the Synch Flag.
**
** Parameters:
**     IN: *hw   - Handle
**         *stat - Pointer to the status variable
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproSynchStatusCPCI(tpropci_dev_t *hw, unsigned char *stat)
{
    unsigned long synch;


    // read the synch status flag bit
    synch = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_STAT_REG) & 0x2;

    // if we are in synch, stat contains the synch source else stat contains 0
    if (synch) 
    {
        *stat = (unsigned char)((PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_STAT_REG) & 0x70000) >> 16);
    }
    else 
    {
        *stat = 0;
    }

    return (0);

} /* End - tproSynchStatusCPCI() */


/*******************************************************************************
**
** Function:    tproPeekCPCI()
** Description: Routine returns the long word value from the selected register.
**
** Parameters:
**     IN: *hw  - Handle
**         *Mem - Pointer to memory object.
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproPeekCPCI(tpropci_dev_t *hw, MemObj *Mem)
{
    Mem->l_value = PEEK_REG_CPCI(hw->ioAddr, Mem->offset);
    Mem->value   = (unsigned short)Mem->l_value;

    return (0);

} /* End - tproPeekCPCI() */


/*******************************************************************************
**
** Function:    tproPokeCPCI()
** Description: Routine pokes the value into the selected register.
**
** Parameters:
**     IN: *hw  - Handle
**         *Mem - Pointer to memory object.
**
**     RETURNS: (0) Success
**
*******************************************************************************/
unsigned char tproPokeCPCI(tpropci_dev_t *hw, MemObj *Mem)
{
    POKE_REG_CPCI(hw->ioAddr, Mem->offset, Mem->l_value);

    return (0);

} /* End - tproPoke() */


/*******************************************************************************
**
** Function:    tproSetOscCPCI()
** Description: Stub function for setting the oscillator output.
**
** Parameters:
**     IN: *hw   - Handle
**         freq  - Pointer to oscillator frequency setting.
**
**     RETURNS: (0) Success
**              (1) Failure
**
*******************************************************************************/
unsigned char tproSetOscCPCI(tpropci_dev_t *hw, unsigned char *freq)
{
    unsigned int command[4];


    // initialize command registers
    command[0] = 0;
    command[1] = 0;
    command[2] = 0;
    command[3] = 0;

    // send command based upon frequency index
    switch (*freq)
    {
        case '0':  // OFF
        default:
            command[3] = 0x45;
            break;

        case '1':  // 1 KHz
            command[3] = 0x145;
            break;

        case '2':  // 1 MHz 
            command[3] = 0x245;
            break;

        case '3':  // 5 MHz
            command[3] = 0x345;
            break;

        case '4':  // 10 MHz
            command[3] = 0x445;
            break;
    }

    // send commands
    sendCmdCPCI(hw->ioAddr, command);

    // verify response
    if ((PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_RESP_REG_3) & 0xFFFF) == 0xFF45)
    {
        return (1);
    }

    return (0);

} /* End - tproSetOscCPCI() */


/*******************************************************************************
**
** Function:    tproISRCPCI()
** Description: interrupt handler
**
** Parameters:
**     IN: *hw   - Handle
**         *irq  - pointer to interrupt
**
**     RETURNS: (0) Interrupt Handled
**              (1) Interrupt Not Handled
**
*******************************************************************************/
unsigned char tproISRCPCI (tpropci_dev_t *hw, int *irq)
{
    unsigned long stat;
    unsigned int  val;


    /* read the interrupt status register */
    stat = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_STAT_REG); 

    /* check if the interrupt is ours */
    /* match int mask bit 8 */
    /* match Flag bit 2 */
    /* heartbeat int mask bit 9 */
    /* heartbeat Flag bit 3 */
    /* Time Tag int mask bit 10 */
    /* Time Tag Flag bit 4 */
    if (((stat & 0x104) != 0x104) && /* match int and match flag*/
        ((stat & 0x208) != 0x208) && /* heartbeat int and heartbeat flag */
        ((stat & 0x410) != 0x410))   /* time tag Int and time tag rdy */
    {
        return (1);
    }

    // match interrupt
    if (stat & 0x4)
    {
        // match condition is met - clear flag
        POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_M, 0x00);

        /* if we are waiting for interrupt post notification */
        if (atomic_read(&hw->matchFlag) == 0)
        {
            atomic_inc(&hw->matchFlag);
            os_waitPost(&hw->match_wait);
        }
    }

    // heartbeat interrupt
    if (stat & 0x8)
    {
        // heartbeat interrupt has occurred - clear flag
        POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_HB, 0x00);

        /* if we are waiting for interrupt post notification */
        if (atomic_read(&hw->heartFlag) == 0)
        {
            atomic_inc(&hw->heartFlag);
            os_waitPost(&hw->heart_wait);
        }
        // heartbeat interrupt has occurred - clear flag
        POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_HB, 0x00);

    }

    // time tag interrupt
    if (stat & 0x10)
    {
        unsigned short *ptr = &hw->fifoVector[hw->headIndex];

        // initialize start of fifo
        ptr[0] = 0x00;
        ptr[1] = 0x00;

        hw->eventCnt++;

        // time tag interrupt has occurred - clear flag
        POKE_REG_CPCI(hw->ioAddr, TPRO_CPCI_CLR_FLAG_TT, 0x00);

        // read status register -- this latches the tag time and date
        val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_STAT);

        // read upper register
        val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_UPPER);

        // parse days
        ptr[2] = (unsigned short)((val & 0x0F000000) >> 24);
        ptr[3] = (unsigned short)((val & 0x00FF0000) >> 16);

        // parse hours
        ptr[4] = (unsigned short)((val & 0x0000FF00) >>  8);

        // parse minutes
        ptr[5] = (unsigned short)((val & 0x000000FF) >>  0);

        // read lower clock register
        val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_LOWER);

        // parse seconds
        ptr[6] = (unsigned short)((val & 0xFF000000) >>  24);
        ptr[7] = (unsigned short)((val & 0x00FF0000) >>  16);
        ptr[8] = (unsigned short)((val & 0x0000FF00) >>   8);
        ptr[9] = (unsigned short)((val & 0x000000FF) >>   0);

        // date dummy read
        val = PEEK_REG_CPCI(hw->ioAddr, TPRO_CPCI_TTAG_DATE);

        /* increment head index */
        hw->headIndex = ((hw->headIndex + 10) % (sizeof(hw->fifoVector)/ sizeof(hw->fifoVector[0])));

        /* if we are waiting for interrupt post notification */
        if (atomic_read(&hw->eventFlag) == 0)
        {
            atomic_inc(&hw->eventFlag);
            os_waitPost(&hw->event_wait);
        }
    }

    hw->intCnt++;

    return (0);

} /* End - tproISRCPCI() */
