/*******************************************************************************
**
**  Module  : tpro.h
**  Date    : 04/05/06
**  Purpose : This is the TPRO-PCI interface include file. 
**
**  Copyright(C) 2006 Spectracom Corporation. All Rights Reserved.
**
*******************************************************************************/

#ifndef _defined_TPRO_
#define _defined_TPRO_

/***************************************************************************
            DEFINES
***************************************************************************/

/*
**  Heartbeat constants
*/
#define SIG_PULSE        (0xE5) /* heartbeat is a pulse */
#define SIG_SQUARE       (0xE7) /* heartbeat is a squarewave */

#define SIG_NO_JAM       (0)    /* start next cycle */
#define SIG_JAM          (1)    /* start immediately */

#define HEART_NORMAL     (0)    // signal type for CPCI/PMC card
#define HEART_INVERT     (8)    // signal type for CPCI/PMC card

#define HEART_DISABLE    (0)    // output type for CPCI/PMC card
#define HEART_ENABLE     (2)    // output type for CPCI/PMC card

#define CLK_10MHZ        (0)    // clock frequency for CPCI/PMC board
#define CLK_3MHZ         (1)    // clock frequency for CPCI/PMC board
#define CLK_1MHZ         (2)    // clock frequency for CPCI/PMC board
#define CLK_1KHZ         (3)    // clock frequency for CPCI/PMC board
/*
**  Match constants
*/
#define MATCH_TIME_START (0) /* start time */
#define MATCH_TIME_STOP  (1) /* stop time */

/*
**  Oscillator frequencies - for Compact PCI Card Only
*/
#define OSC_OUT_OFF      (0)
#define OSC_OUT_1KHZ     (1)
#define OSC_OUT_1MHZ     (2)
#define OSC_OUT_5MHZ     (3)
#define OSC_OUT_10MHZ    (4)
 
/*
** TPRO BOARD OBJECT
*/
typedef struct TPRO_BoardObj {

  int            file_descriptor;
  unsigned short devid;
  unsigned short options;

} TPRO_BoardObj;

/*
** TPRO ALTITUDE OBJECT
*/
typedef struct TPRO_AltObj { 

  float meters;

} TPRO_AltObj;

/*
** TPRO DATE OBJECT
*/
typedef struct TPRO_DateObj {

  unsigned short year;
  unsigned char  month;
  unsigned char  day;

} TPRO_DateObj;

/*
** TPRO LONGITUDE/LATTITUDE OBJECT
*/
typedef struct TPRO_LongLat { 

  unsigned short degrees;
  float          minutes;

} TPRO_LongObj, TPRO_LatObj;

/*
** TPRO MATCH OBJECT
*/
typedef struct TPRO_MatchObj { 

  unsigned char   matchType; /* start/stop time */
  double          seconds;
  unsigned char   minutes;
  unsigned char   hours;
  unsigned short  days;

} TPRO_MatchObj;

/*
** TPRO SATINFO OBJECT
*/
typedef struct TPRO_SatObj {

  unsigned char satsTracked; /* num sats tracked */
  unsigned char satsView;    /* num sats in view */

} TPRO_SatObj;

/*
** TPRO HEARTBEAT OBJECT
*/
typedef struct TPRO_HeartObj { 

  unsigned char signalType; // square or pulse  -- ignored by cPCI/PMC
  unsigned char outputType; // jamming option   (or clock select for cPCI/PMC)
  double        frequency;  // heartbeat freq

} TPRO_HeartObj;

/*
** TPRO TIME OBJECT
*/
typedef struct TPRO_TimeObj { 

  double         secsDouble; /* seconds floating pt */
  unsigned char  seconds;    /* seconds whole num */
  unsigned char  minutes;
  unsigned char  hours;
  unsigned short days;
  unsigned short year;
  unsigned short flags;      /* bit 15 flagsInvalid(1); bit 2 SYNC, bit 1 TCODE; all others 0 */

} TPRO_TimeObj;

/*
** TPRO WAIT OBJECT
*/
typedef struct TPRO_WaitObj {

  int jiffies;            /*-- # jiffies to wait ---*/
  double  seconds;
  unsigned char  minutes;
  unsigned char  hours;
  unsigned short days;

} TPRO_WaitObj;

/*
** TPRO MEM OBJECT FOR PEEK/POKE
*/
typedef struct TPRO_MemObj { 

  unsigned short offset;
  unsigned short value;
  unsigned long  l_value;

} TPRO_MemObj;

/*******************************************************************************
            ERROR CODES
*******************************************************************************/
#define TPRO_SUCCESS                (0)  /* success */
#define TPRO_HANDLE_ERR             (1)  /* error bad handle */
#define TPRO_OBJECT_ERR             (2)  /* error creating obj */
#define TPRO_CLOSE_HANDLE_ERR       (3)  /* err closing device */
#define TPRO_DEVICE_NOT_OPEN_ERR    (4)  /* device not opened */
#define TPRO_INVALID_BOARD_TYPE_ERR (5)  /* invalid device */
#define TPRO_FREQ_ERR               (6)  /* invalid frequency */
#define TPRO_YEAR_PARM_ERR          (7)  /* invalid year */
#define TPRO_DAY_PARM_ERR           (8)  /* invalid day */
#define TPRO_HOUR_PARM_ERR          (9)  /* invalid hour */
#define TPRO_MIN_PARM_ERR          (10)  /* invalid minutes */
#define TPRO_SEC_PARM_ERR          (11)  /* invalid seconds */
#define TPRO_DELAY_PARM_ERR        (12)  /* invalid delay */
#define TPRO_TIMEOUT_ERR           (13)  /* device timed out */
#define TPRO_COMM_ERR              (14)  /* communication error */

/*******************************************************************************
            PUBLIC ROUTINE PROTOTYPES
*******************************************************************************/
unsigned char TPRO_open             (TPRO_BoardObj **hnd, char *deviceName);
unsigned char TPRO_close            (TPRO_BoardObj *hnd);
unsigned char TPRO_getAltitude      (TPRO_BoardObj *hnd, TPRO_AltObj *Altp);
unsigned char TPRO_getDate          (TPRO_BoardObj *hnd, TPRO_DateObj *Datep);
unsigned char TPRO_getDriver        (TPRO_BoardObj *hnd, char *driver);
unsigned char TPRO_getFirmware      (TPRO_BoardObj *hnd, char *firmware);
unsigned char TPRO_getFPGA          (TPRO_BoardObj *hnd, char *fpga);
unsigned char TPRO_getLatitude      (TPRO_BoardObj *hnd, TPRO_LatObj *Latp);
unsigned char TPRO_getLongitude     (TPRO_BoardObj *hnd, TPRO_LongObj *Longp);
unsigned char TPRO_getSatInfo       (TPRO_BoardObj *hnd, TPRO_SatObj *Satp);
unsigned char TPRO_getTime          (TPRO_BoardObj *hnd, TPRO_TimeObj *Timep);
unsigned char TPRO_resetFirmware    (TPRO_BoardObj *hnd);
unsigned char TPRO_setHeartbeat     (TPRO_BoardObj *hnd, TPRO_HeartObj *Heartp);
unsigned char TPRO_setMatchTime     (TPRO_BoardObj *hnd, TPRO_MatchObj *Matchp);
unsigned char TPRO_setOscillator    (TPRO_BoardObj *hnd, unsigned char *freq);
unsigned char TPRO_setPropDelayCorr (TPRO_BoardObj *hnd, int *us);
unsigned char TPRO_setTime          (TPRO_BoardObj *hnd, TPRO_TimeObj *Timep);
unsigned char TPRO_setYear          (TPRO_BoardObj *hnd, unsigned short *yr);
unsigned char TPRO_simEvent         (TPRO_BoardObj *hnd);
unsigned char TPRO_synchControl     (TPRO_BoardObj *hnd, unsigned char *enbp);
unsigned char TPRO_synchStatus      (TPRO_BoardObj *hnd, unsigned char *status);
unsigned char TPRO_waitEvent        (TPRO_BoardObj *hnd, TPRO_WaitObj *waitp);
unsigned char TPRO_waitHeartbeat    (TPRO_BoardObj *hnd, int *jiffies);
unsigned char TPRO_waitMatch        (TPRO_BoardObj *hnd, int *jiffies);
unsigned char TPRO_peek             (TPRO_BoardObj *hnd, TPRO_MemObj *pMem);
unsigned char TPRO_poke             (TPRO_BoardObj *hnd, TPRO_MemObj *pMem);

/*******************************************************************************
            PUBLIC ROUTINE AVAILABILITY
*******************************************************************************/

/*                                                             Available only to
                       Available when  Available to all users  first user when
Routine                #users = 1      when #users > 1         #users > 1
---------------------  --------------  ----------------------  ----------------- 
TPRO_open                     Y                   Y                             
TPRO_close                    Y                   Y                             
TPRO_getAltitude              Y                                        Y        
TPRO_getDate                  Y                                        Y        
TPRO_getDriver                Y                   Y                            
TPRO_getFirmware              Y                                        Y        
TPRO_getFPGA                  Y                                        Y        
TPRO_getLatitude              Y                                        Y        
TPRO_getLongitude             Y                                        Y        
TPRO_getSatInfo               Y                                        Y        
TPRO_getTime                  Y                   Y                             
TPRO_resetFirmware            Y                                        Y        
TPRO_setHeartbeat             Y                                        Y        
TPRO_setMatchTime             Y                                        Y        
TPRO_setOscillator            Y                                        Y        
TPRO_setPropDelayCorr         Y                                        Y        
TPRO_setTime                  Y                                        Y        
TPRO_setYear                  Y                                        Y        
TPRO_simEvent                 Y                                        Y        
TPRO_synchControl             Y                                        Y        
TPRO_synchStatus              Y                   Y                             
TPRO_waitEvent                Y                                        Y        
TPRO_waitHeartbeat            Y                                        Y        
TPRO_waitMatch                Y                                        Y        
TPRO_peek                     Y                   Y                             
TPRO_poke                     Y                   Y                             
*/

#endif	// _defined_TPRO_
