/*******************************************************************************
 *
 *  Module  : refclock_tpropci.c
 *  Date    : 04/17/06
 *  Purpose : Implements a reference clock driver for the NTP daemon.  This
 *            reference clock driver provides a means to communicate with 
 *            the Spectracom TPRO/TSAT PCI timing devices and use them as a time 
 *            source. 
 *
 ******************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(REFCLOCK) && defined(CLOCK_TPROPCI)

#include <asm/ioctl.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <ctype.h>

#include "ntpd.h"
#include "ntp_io.h"
#include "ntp_refclock.h"
#include "ntp_unixtime.h"
#include "ntp_stdlib.h"


/*******************************************************************************
**
** This driver supports the KSI TPRO/TSAT-PCI GPS receiver.  It requires
** that the tpropci.o device driver be installed and loaded.
** 
*******************************************************************************/

#define TSAT_PCI_REVISION "1.20"

/*
** TPRO interface definitions
*/
#define DEVICE      "/dev/tpropci"              /* device name */
#define PRECISION   (-20)                       /* precision assumed (1 us) */
#define REFID       "GPS/"                      /* reference ID */
#define DESCRIPTION "Spectracom TPRO/TSAT-PCI"  /* WRU */

#define SECONDS_1900_TO_1970 (2208988800U)

/*
**  Function prototypes
*/
static void tpro_poll     (int unit, struct peer *);
static void tpro_shutdown (int, struct peer *);
static int  tpro_start    (int, struct peer *);

/*
**  Transfer vector
*/
struct refclock refclock_tpropci = {
    tpro_start,    /* start up driver */
    tpro_shutdown, /* shut down driver */
    tpro_poll,     /* transmit poll message */
    noentry,       /* not used (old tpro_control) */
    noentry,       /* initialize driver (not used) */
    noentry,       /* not used (old tpro_buginfo) */
    NOFLAGS        /* not used */
};

/*
** TPRO/TSAT Board Object
*/
typedef struct BoardObj {
 
  int            file_descriptor;
  unsigned short devid;
  unsigned short options;
  unsigned char  firmware[5];
  unsigned char  FPGA[5];
  unsigned char  driver[7];

}BoardObj;

/*
** TPRO/TSAT Time Object
*/
typedef struct TimeObj {
 
  unsigned char  syncOption; /* -M option */
  unsigned int   secsDouble; /* seconds floating pt */
  unsigned char  seconds;    /* seconds whole num */
  unsigned char  minutes;    /* minutes */
  unsigned char  hours;      /* hours */
  unsigned short days;       /* days */
  unsigned short year;       /* year for CPCI board */

} TimeObj;

/*
** NTP Time Object
*/
typedef struct NtpTimeObj {

    TimeObj        timeObj;
    struct timeval tv;
    unsigned int   refId;

} NtpTimeObj;

/*******************************************************************************
**          IOCTL DEFINITIONS
*******************************************************************************/
#define IOCTL_TPRO_ID            't'
#define IOCTL_TPRO_OPEN          _IOWR(IOCTL_TPRO_ID, 0, BoardObj)
#define IOCTL_TPRO_GET_NTP_TIME  _IOWR(IOCTL_TPRO_ID, 25, NtpTimeObj)
#define IOCTL_TPRO_SYNCH_STATUS  _IOWR(IOCTL_TPRO_ID, 19, unsigned char *)

/******************************************************************************
 *
 * Function:    tpro_start()
 * Description: Used to intialize the Spectracom TPRO/TSAT reference driver.
 *
 * Parameters:
 *     IN:  unit - not used.
 *         *peer - pointer to this reference clock's peer structure      
 *     Returns: 0 - unsuccessful
 *              1 - successful 
 *
*******************************************************************************/
static int tpro_start(int unit, struct peer *peer)
{
    struct refclockproc *pp;

    /*
    **  initialize reference clock and peer parameters
    */
    pp                = peer->procptr;
    pp->io.clock_recv = noentry;
    pp->io.srcclock   = (caddr_t)peer;
    pp->io.datalen    = 0;
    peer->precision   = PRECISION;
//    peer->burst       = NSTAGE;
    pp->clockdesc     = DESCRIPTION;
    
    memcpy((char *)&pp->refid, REFID, 4);

    return (1);

} /* End - tpro_start() */

/*******************************************************************************
**
** Function:    tpro_shutdown()
** Description: Handles anything related to shuting down the reference clock
**              driver. Nothing at this point in time.
**
** Parameters:
**     IN:  unit - not used.
**         *peer - pointer to this reference clock's peer structure      
**     Returns: none. 
**
*******************************************************************************/
static void tpro_shutdown(int unit, struct peer *peer)
{

} /* End - tpro_shutdown() */

/******************************************************************************
 *
 * Function:    tpro_poll()
 * Description: Retrieve time from the TPRO/TSAT device.
 *
 * Parameters:
 *     IN:  unit - not used.
 *         *peer - pointer to this reference clock's peer structure      
 *     Returns: none. 
 *
*******************************************************************************/
static void tpro_poll(int unit, struct peer *peer)
{
    char                 device[32];
    struct refclockproc *pp;
    unsigned char        synch;
    double               seconds;
    int                  err;  
    NtpTimeObj           TimeContext;
    BoardObj             hBoard;

    /* Construct the device name */
    sprintf(device, "%s%d", DEVICE, (int)peer->refclkunit);
    
    printf("Pollind device number %d...\n", (int)peer->refclkunit);
    
    /* Open the TPRO/TSAT device */
    hBoard.file_descriptor = open(device, O_RDONLY | O_NDELAY, 0777);
    
    /* If error opening TPRO/TSAT device... */
    if (hBoard.file_descriptor < 0) 
    {
        msyslog(LOG_ERR, "Couldn't open device\n");
        return;
    }

    /* If error while initializing the board... */
    if (ioctl(hBoard.file_descriptor, IOCTL_TPRO_OPEN, &hBoard) < 0) 
    {
        msyslog(LOG_ERR, "Couldn't initialize device\n");
        close(hBoard.file_descriptor);
        return;
    }

    pp = peer->procptr;

    /* See if we are synchronized */
    err = ioctl(hBoard.file_descriptor, IOCTL_TPRO_SYNCH_STATUS, &synch);
    
    if (err < 0) {
        refclock_report(peer, CEVNT_FAULT);
        close(hBoard.file_descriptor);
        return;
    }

    /* Read the time from the TPRO/TSAT-PCI device */
    err = ioctl(hBoard.file_descriptor, 
                IOCTL_TPRO_GET_NTP_TIME, 
                (char *)&TimeContext);
                
    /* Close the TPRO/TSAT device */
    close(hBoard.file_descriptor);

    if (err < 0) {
        refclock_report(peer, CEVNT_FAULT);
        return;
    }

    seconds = (double) (TimeContext.timeObj.secsDouble);
    seconds /= (double) 1000000.0;

    /*
    ** Convert the number of microseconds to double and then place in the
    ** peer's last received long floating point format.
    */ 
    DTOLFP(((double)TimeContext.tv.tv_usec / 1000000.0), &pp->lastrec);
    
    /*
    ** The specTimeStamp is the number of seconds since 1/1/1970, while the 
    ** peer's lastrec time should be compatible with NTP which is seconds since
    ** 1/1/1900.  So Add the number of seconds between 1900 and 1970 to the
    ** specTimeStamp and place in the peer's lastrec long floating point struct.
    */
    pp->lastrec.Ul_i.Xl_ui += (unsigned int)TimeContext.tv.tv_sec + 
                                            SECONDS_1900_TO_1970;    
    
    pp->polls++;

    /*
    **  set the reference clock object
    */
    sprintf(pp->a_lastcode, "%03d %02d:%02d:%02.6f", TimeContext.timeObj.days, 
                                                     TimeContext.timeObj.hours, 
                                                     TimeContext.timeObj.minutes, 
                                                     seconds);
    
    pp->lencode = strlen (pp->a_lastcode);
    pp->day     = TimeContext.timeObj.days;
    pp->hour    = TimeContext.timeObj.hours;
    pp->minute  = TimeContext.timeObj.minutes;
    pp->second  = (int) seconds;
    seconds     = (seconds - (double) (pp->second / 1.0)) * 1000000000;
    pp->nsec    = (long) seconds;

    if (synch) {
        pp->leap = LEAP_NOWARNING;
    }
    else {
        pp->leap = LEAP_NOTINSYNC;
        refclock_report(peer, CEVNT_FAULT);
        return;
    }

    if (!refclock_process(pp)) {
        refclock_report(peer, CEVNT_BADTIME);
        return;
    }

/*    if (peer->burst > 0) {
        return;
    }
*/
    if (pp->coderecv == pp->codeproc) {
        refclock_report(peer, CEVNT_TIMEOUT);
        return;
    }

    record_clock_stats(&peer->srcadr, pp->a_lastcode);
    refclock_receive(peer);

    /* Reset for the next burst of polls */
//    peer->burst = NSTAGE;

    /* Increment the number of times the refernece has been polled */
	pp->polls++;

} /* End - tpro_poll() */

#else
int refclock_tpropci_bs;
#endif /* REFCLOCK */
